from django import forms
from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.forms import ReadOnlyPasswordHashField
from django.contrib.auth.models import Group as BaseGroup
from django.contrib.contenttypes.admin import GenericTabularInline
from django.db.models.functions import Collate
from enumfields.admin import EnumFieldListFilter

from arkindex.users.models import Group, Right, User, UserScope


class UserCreationForm(forms.ModelForm):
    """A form for creating new users. Includes all the required
    fields, plus a repeated password."""
    password1 = forms.CharField(label="Password", widget=forms.PasswordInput)
    password2 = forms.CharField(label="Password confirmation", widget=forms.PasswordInput)

    class Meta:
        model = User
        fields = ("email", )

    def clean_password2(self):
        # Check that the two password entries match
        password1 = self.cleaned_data.get("password1")
        password2 = self.cleaned_data.get("password2")
        if password1 and password2 and password1 != password2:
            raise forms.ValidationError("Passwords don't match")
        return password2

    def save(self, commit=True):
        # Save the provided password in hashed format
        user = super().save(commit=False)
        user.set_password(self.cleaned_data["password1"])
        if commit:
            user.save()
        return user


class UserChangeForm(forms.ModelForm):
    """A form for updating users. Includes all the fields on
    the user, but replaces the password field with admin's
    password hash display field.
    """
    password = ReadOnlyPasswordHashField(help_text='<a href="../password/">Change password</a>')

    class Meta:
        model = User
        fields = ("email", "password", "is_active", "is_admin")

    def clean_password(self):
        # Regardless of what the user provides, return the initial value.
        # This is done here, rather than on the field, because the
        # field does not have access to the initial value
        return self.initial["password"]


class UserAdmin(BaseUserAdmin):
    # The forms to add and change user instances
    form = UserChangeForm
    add_form = UserCreationForm

    # The fields to be used in displaying the User model.
    # These override the definitions on the base UserAdmin
    # that reference specific fields on auth.User.
    list_display = ("email", "display_name", "is_admin", "created")
    list_filter = ("is_admin", )
    fieldsets = (
        (None, {"fields": ("email", "display_name", "verified_email", "password")}),
        ("Permissions", {"fields": ("is_admin", "is_active")}),
    )
    # add_fieldsets is not a standard ModelAdmin attribute. UserAdmin
    # overrides get_fieldsets to use this attribute when creating a user.
    add_fieldsets = (
        (None, {
            "classes": ("wide",),
            "fields": ("email", "password1", "password2")}
         ),
    )
    search_fields = ("searchable_email",)
    ordering = ("-created", "email",)
    filter_horizontal = ()

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        # Django's search uses ILIKE, but LIKE is not supported in PostgreSQL with nondeterministic collations.
        # We are using a case-insensitive collation to get unique case-insensitive emails, so to search properly,
        # we first have to set the collation back to its default value, making it case-sensitive again;
        # since Django uses ILIKE, it will be case-insensitive anyway.
        return qs.annotate(searchable_email=Collate("email", "default"))


class UserScopeAdmin(admin.ModelAdmin):
    list_display = ("user", "scope")
    list_filter = [("scope", EnumFieldListFilter), ]


class GroupMembershipInline(GenericTabularInline):
    ct_fk_field = "content_id"
    model = Right
    fields = ("group", "level")
    extra = 1

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.filter(user__isnull=True)


class UserMembershipInline(GenericTabularInline):
    ct_fk_field = "content_id"
    model = Right
    fields = ("user", "level")
    extra = 1

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.filter(group__isnull=True)


class GroupAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "public")
    inlines = (UserMembershipInline, )


admin.site.register(User, UserAdmin)
# Register the custom GroupAdmin
admin.site.register(Group, GroupAdmin)
# and hide base GroupAdmin form contrib.auth
admin.site.unregister(BaseGroup)
admin.site.register(UserScope, UserScopeAdmin)
